/*
 *  Julia.cpp - Realtime Julia/Mandelbrot set V3.0
 *              Written by Christian Bauer
 *              Public domain
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>


const ulong APP_SIGNATURE = '????';

// Window/Application messages
const ulong MSG_NEW_WINDOW = 'neww';
const ulong MSG_RESIZE = 'size';
const ulong MSG_MOUSE_DOWN = 'mdw1';
const ulong MSG_MOUSE_DOWN_2 = 'mdw2';
const ulong MSG_MOUSE_DOWN_3 = 'mdw3';
const ulong MSG_MOUSE_MOVED = 'mmov';
const ulong MSG_ITER_16 = 'i16 ';
const ulong MSG_ITER_32 = 'i32 ';
const ulong MSG_ITER_64 = 'i64 ';
const ulong MSG_ITER_128 = 'i128';
const ulong MSG_ITER_256 = 'i256';
const ulong MSG_TYPE_JULIA = 'jlia';
const ulong MSG_TYPE_MANDEL = 'mand';

// Maximum window size
const int MAX_SIZE_X = 1024;
const int MAX_SIZE_Y = 1024;

// Types
typedef unsigned char UBYTE;

// Number of CPUs in this system
int CPUCount;


class JuliaWindow;
class BitmapView;


// Data structure to pass parameters to threads
struct thread_data {
	UBYTE *bits;
	int xmod;
	int size_x, size_y;
	double cx, cy;
	double xs, ys;
	double x_factor, y_factor;
	int max_iter, starty, skipy;
};


// Application object
class Julia : public BApplication {
public:
	Julia();
	virtual void ReadyToRun(void);
	virtual void AboutRequested(void);
	virtual void MessageReceived(BMessage *msg);
};


// Window object
class JuliaWindow : public BWindow {
public:
	JuliaWindow();
	virtual bool QuitRequested(void);
	virtual void MessageReceived(BMessage *msg);

private:
	void reset_params(void);
	void recalc(void);
	static void recalc_julia(UBYTE *bits, int xmod, int size_x, int size_y, double cx, double cy, double x_factor, double y_factor, int max_iter, int starty, int skipy);
	static void recalc_mandel(UBYTE *bits, int xmod, int size_x, int size_y, double cx, double cy, double x_factor, double y_factor, int max_iter, int starty, int skipy);
	static long thread_func_julia(void *arg);
	static long thread_func_mandel(void *arg);

	BBitmap *the_bitmap;
	int xmod;
	BitmapView *main_view;

	// Number of CPUs and parameter passing structures
	int cpu_count;
	thread_data tdata[B_MAX_CPU_COUNT];

	// Drawing parameters
	double cx, cy;				// Iteration constant for Julia set
	double xs, ys;				// top-left X/Y coordinates for Mandelbrot set
	double x_factor, y_factor;	// These are for mapping pixels to coordinates
	int max_iter;				// Maximum iteration depth
	int size_x, size_y;			// Window size in pixels
	bool is_mandel;

	// For Mandelbrot drag/zoom
	long drag_start_x, drag_start_y;
};


// Bitmap view
class BitmapView : public BView {
public:
	BitmapView(BRect frame, BBitmap *bitmap);
	virtual void Draw(BRect update);
	virtual void FrameResized(float width, float height);
	virtual void MouseDown(BPoint point);
	virtual void MouseMoved(BPoint where, ulong code, BMessage *a_message);

private:
	BBitmap *the_bitmap;
};


// Create application object and start it
int main(int argc, char **argv)
{	
	Julia *the_app;

	the_app = new Julia();
	the_app->Run();
	delete the_app;

	return 0;
}


/*
 *  Application constructor, initialize member variables
 */

Julia::Julia() : BApplication(APP_SIGNATURE)
{
	// Determine number of CPUs
	system_info sysinfo;
	get_system_info(&sysinfo);
	CPUCount = sysinfo.cpu_count;

}


/*
 *  Arguments processed, open window
 */

void Julia::ReadyToRun(void)
{
	// Create application main menu
	BPopUpMenu *menu = new BPopUpMenu("", FALSE, FALSE);
	menu->AddItem(new BMenuItem("About Julia...", new BMessage(B_ABOUT_REQUESTED)));
	menu->AddItem(new BSeparatorItem());
	menu->AddItem(new BMenuItem("New Window", new BMessage(MSG_NEW_WINDOW), 'N'));
	menu->AddItem(new BSeparatorItem());
	menu->AddItem(new BMenuItem("Quit Julia", new BMessage(B_QUIT_REQUESTED), 'Q'));
	menu->SetTargetForItems(be_app);
	SetMainMenu(menu);

	// Open first window
	new JuliaWindow;
}


/*
 *  About requested
 */

void Julia::AboutRequested(void)
{
	BAlert *the_alert = new BAlert("",
		"Realtime fractals by Christian Bauer\n"
		"<cbauer@iphcip1.physik.uni-mainz.de>\n"
		"Public domain.\n\n"
		"Julia mode:\n"
		"Mouse button 1 to change iteration constant\n\n"
		"Mandelbrot mode:\n"
		"Mouse button 1 to move\n"
		"Mouse button 2 to zoom in\n"
		"Mouse button 3 to zoom out\n",
		"Neat");
	the_alert->Go();
}


/*
 *  Handle messages
 */

void Julia::MessageReceived(BMessage *msg)
{
	switch (msg->what) {

		case MSG_NEW_WINDOW:	// Open new window
			new JuliaWindow;
			break;

		default:
			inherited::MessageReceived(msg);
	}
}


/*
 *  Window constructor
 */

JuliaWindow::JuliaWindow() : BWindow(BRect(0, 0, 320-1, 240-1), "Realtime Fractals", B_TITLED_WINDOW, 0)
{
	// Move window to right position
	Lock();
	MoveTo(80, 60);

	// Set up menus
	BMenuBar *bar = new BMenuBar(Bounds(), "");
	BMenu *menu = new BMenu("Iterations");
	menu->AddItem(new BMenuItem("16", new BMessage(MSG_ITER_16)));
	menu->AddItem(new BMenuItem("32", new BMessage(MSG_ITER_32)));
	menu->AddItem(new BMenuItem("64", new BMessage(MSG_ITER_64)));
	menu->AddItem(new BMenuItem("128", new BMessage(MSG_ITER_128)));
	menu->AddItem(new BMenuItem("256", new BMessage(MSG_ITER_256)));
	menu->SetRadioMode(TRUE);
	menu->FindItem(MSG_ITER_64)->SetMarked(TRUE);
	bar->AddItem(menu);
	menu = new BMenu("Type");
	menu->AddItem(new BMenuItem("Julia", new BMessage(MSG_TYPE_JULIA)));
	menu->AddItem(new BMenuItem("Mandelbrot", new BMessage(MSG_TYPE_MANDEL)));
	menu->SetRadioMode(TRUE);
	menu->FindItem(MSG_TYPE_JULIA)->SetMarked(TRUE);
	bar->AddItem(menu);
	AddChild(bar);
	SetKeyMenuBar(bar);
	int mbar_height = bar->Frame().bottom + 1;

	// Resize window to fit menu bar
	ResizeBy(0, mbar_height);
	SetSizeLimits(32, MAX_SIZE_X, mbar_height+32, MAX_SIZE_Y);
	SetDiscipline(FALSE);

	// Allocate bitmap
	the_bitmap = new BBitmap(BRect(0, 0, MAX_SIZE_X-1, MAX_SIZE_Y-1), B_COLOR_8_BIT);
	xmod = the_bitmap->BytesPerRow();

	// Set initial parameters
	size_x = 320;
	size_y = 240;
	reset_params();
	max_iter = 64;
	is_mandel = FALSE;
	drag_start_x = drag_start_y = 0;

	// Set constant data for drawing threads
	for (int i=0; i<CPUCount; i++) {
		tdata[i].bits = (UBYTE *)the_bitmap->Bits();
		tdata[i].xmod = xmod;
		tdata[i].starty = i;
		tdata[i].skipy = CPUCount;
	}

	// Create bitmap view
	main_view = new BitmapView(BRect(0, mbar_height, size_x-1, size_y+mbar_height-1), the_bitmap);
	AddChild(main_view);

	// Draw initial julia set
	recalc();

	// Show the window
	Unlock();
	Show();
}


/*
 *  Closing the last window quits the program
 */

bool JuliaWindow::QuitRequested(void)
{
	if (be_app->CountWindows() == 1)
		be_app->PostMessage(B_QUIT_REQUESTED);

	return TRUE;
}


/*
 *  Handle messages
 */

void JuliaWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {

		case MSG_MOUSE_DOWN:	// Primary mouse button
			if (is_mandel) {
				// Save coordinates
				drag_start_x = msg->FindLong("x");
				drag_start_y = msg->FindLong("y");
			} else {
				// Convert mouse coordinates to iteration constant
				cx = double(msg->FindLong("x") - size_x/2) * x_factor;
				cy = double(msg->FindLong("y") - size_y/2) * y_factor;

				// Redraw Julia set
				recalc();
			}
			break;

		case MSG_MOUSE_DOWN_2:	// Secondary mouse button
			if (is_mandel) {
				// Zoom in
				double old_x_factor = x_factor, old_y_factor = y_factor;
				x_factor *= 0.9;
				y_factor *= 0.9;
				xs = ((xs + msg->FindLong("x") * old_x_factor) / x_factor - msg->FindLong("x")) * x_factor;
				ys = ((ys + msg->FindLong("y") * old_y_factor) / y_factor - msg->FindLong("y")) * y_factor;
				recalc();

				// Mouse still down? Then repeat
				BPoint point;
				ulong buttons;
				main_view->GetMouse(&point, &buttons);
				point.ConstrainTo(main_view->Bounds());
				if (buttons & B_SECONDARY_MOUSE_BUTTON) {
					BMessage *newmsg = new BMessage(MSG_MOUSE_DOWN_2);
					newmsg->AddLong("x", point.x);
					newmsg->AddLong("y", point.y);
					PostMessage(newmsg);
				}
			}
			break;

		case MSG_MOUSE_DOWN_3:	// Tertiary mouse button
			if (is_mandel) {
				// Zoom out
				double old_x_factor = x_factor, old_y_factor = y_factor;
				x_factor *= 1.1;
				y_factor *= 1.1;
				xs = ((xs + msg->FindLong("x") * old_x_factor) / x_factor - msg->FindLong("x")) * x_factor;
				ys = ((ys + msg->FindLong("y") * old_y_factor) / y_factor - msg->FindLong("y")) * y_factor;
				recalc();

				// Mouse still down? Then repeat
				BPoint point;
				ulong buttons;
				main_view->GetMouse(&point, &buttons);
				point.ConstrainTo(main_view->Bounds());
				if (buttons & B_TERTIARY_MOUSE_BUTTON) {
					BMessage *newmsg = new BMessage(MSG_MOUSE_DOWN_3);
					newmsg->AddLong("x", point.x);
					newmsg->AddLong("y", point.y);
					PostMessage(newmsg);
				}
			}
			break;

		case MSG_MOUSE_MOVED:	// Mouse moved while mouse button down

			// Prevent messages from piling up
			BMessage *msg2;
			MessageQueue()->Lock();
			while ((msg2 = MessageQueue()->FindMessage(MSG_MOUSE_MOVED, 0)) != NULL)
				MessageQueue()->RemoveMessage(msg2);
			MessageQueue()->Unlock();

			if (is_mandel) {
				// Move around complex plane
				xs -= (msg->FindLong("x") - drag_start_x) * x_factor;
				ys -= (msg->FindLong("y") - drag_start_y) * y_factor;
				drag_start_x = msg->FindLong("x");
				drag_start_y = msg->FindLong("y");
			} else {
				// Convert mouse coordinates to iteration constant
				cx = double(msg->FindLong("x") - size_x/2) * x_factor;
				cy = double(msg->FindLong("y") - size_y/2) * y_factor;
			}

			// Redraw Julia/Mandelbrot set
			recalc();
			break;

		case MSG_RESIZE:		// Window was resized

			// Save old size
			long old_size_x = size_x, old_size_y = size_y;

			// Get new size
			size_x = msg->FindLong("x");
			size_y = msg->FindLong("y");

			if (is_mandel) {
				// Calculate new top/left coordinate so that the
				// center remains the same
				xs += (old_size_x - size_x) / 2 * x_factor;
				ys += (old_size_y - size_y) / 2 * y_factor;
			} else {
				// Calculate x and y factor
				if (size_x > size_y) {
					x_factor = 4.0 / double(size_y);
					y_factor = 4.0 / double(size_y);
				} else {
					x_factor = 4.0 / double(size_x);
					y_factor = 4.0 / double(size_x);
				}
			}

			// Redraw Julia/Mandelbrot set
			recalc();
			break;

		case MSG_ITER_16:		// Change iteration depth
			max_iter = 16;
			recalc();
			break;

		case MSG_ITER_32:
			max_iter = 32;
			recalc();
			break;

		case MSG_ITER_64:
			max_iter = 64;
			recalc();
			break;

		case MSG_ITER_128:
			max_iter = 128;
			recalc();
			break;

		case MSG_ITER_256:
			max_iter = 256;
			recalc();
			break;

		case MSG_TYPE_JULIA:	// Change fractal type
			is_mandel = FALSE;
			reset_params();
			recalc();
			break;

		case MSG_TYPE_MANDEL:
			is_mandel = TRUE;
			reset_params();
			recalc();
			break;

		default:
			inherited::MessageReceived(msg);
	}
}


/*
 *  Reset drawing parameters
 */

void JuliaWindow::reset_params(void)
{
	cx = cy = 0.0;
	xs = ys = -2.0;

	if (size_x > size_y) {
		x_factor = 4.0 / double(size_y);
		y_factor = 4.0 / double(size_y);
	} else {
		x_factor = 4.0 / double(size_x);
		y_factor = 4.0 / double(size_x);
	}
}


/*
 *  Recalc and redraw Julia/Mandelbrot set (multithreaded)
 */

void JuliaWindow::recalc(void)
{
	thread_id thread[B_MAX_CPU_COUNT];
	int i;

	// Start as many threads as there are CPUs
	for (i=0; i<CPUCount; i++) {
		tdata[i].size_x = size_x;
		tdata[i].size_y = size_y;
		tdata[i].cx = cx;
		tdata[i].cy = cy;
		tdata[i].xs = xs;
		tdata[i].ys = ys;
		tdata[i].x_factor = x_factor;
		tdata[i].y_factor = y_factor;
		tdata[i].max_iter = max_iter;
		if (is_mandel)
			thread[i] = spawn_thread(thread_func_mandel, "Julia Calc", B_NORMAL_PRIORITY, &tdata[i]);
		else
			thread[i] = spawn_thread(thread_func_julia, "Julia Calc", B_NORMAL_PRIORITY, &tdata[i]);
		resume_thread(thread[i]);
	}

	// Wait for threads to finish
	for (i=0; i<CPUCount; i++) {
		long l;
		wait_for_thread(thread[i], &l);
	}

	// Redraw bitmap
	Lock();
	main_view->Draw(main_view->Bounds());
	Unlock();
}

inline void JuliaWindow::recalc_julia(UBYTE *bits, int xmod, int size_x, int size_y, double cx, double cy, double x_factor, double y_factor, int max_iter, int starty, int skipy)
{
	UBYTE *p = bits + starty * xmod;
	UBYTE *q = bits + (size_y-starty-1) * xmod;

	double ys = -double(size_y/2 - starty) * y_factor;
	double y_add = y_factor * double(skipy);
	for (int iy=starty; iy<=size_y/2; iy+=skipy, ys+=y_add) {

		double xs = -double(size_x/2) * x_factor;
		for (int ix=0; ix<size_x; ix++, xs+=x_factor) {

			double x = xs, y = ys;
			double x2 = x * x, y2 = y * y;
			int iter = 0;
			do {
				iter++;
				y = 2.0 * x * y - cy;
				x = x2 - y2 - cx;
				x2 = x * x;
				y2 = y * y;
			} while ((iter < max_iter) && (x2 + y2 < 4.0));

			if (iter == max_iter)
				p[ix] = q[size_x - ix - 1] = 0;
			else
				p[ix] = q[size_x - ix - 1] = iter + 32;
		}
		p += xmod * skipy;
		q -= xmod * skipy;
	}
}

inline void JuliaWindow::recalc_mandel(UBYTE *bits, int xmod, int size_x, int size_y, double xs, double ys, double x_factor, double y_factor, int max_iter, int starty, int skipy)
{
	UBYTE *p = bits + starty * xmod;
	double x_left = xs;

	ys += starty * y_factor;
	double y_add = y_factor * double(skipy);
	for (int iy=starty; iy<size_y; iy+=skipy, ys+=y_add) {

		double xs = x_left;
		for (int ix=0; ix<size_x; ix++, xs+=x_factor) {

			double x = 0.0, y = 0.0;
			double x2 = x * x, y2 = y * y;
			int iter = 0;
			do {
				iter++;
				y = 2.0 * x * y - ys;
				x = x2 - y2 - xs;
				x2 = x * x;
				y2 = y * y;
			} while ((iter < max_iter) && (x2 + y2 < 4.0));

			if (iter == max_iter)
				p[ix] = 0;
			else
				p[ix] = iter + 32;
		}
		p += xmod * skipy;
	}
}

long JuliaWindow::thread_func_julia(void *arg)
{
	thread_data *data = (thread_data *)arg;
	recalc_julia(data->bits, data->xmod, data->size_x, data->size_y,
				data->cx, data->cy, data->x_factor, data->y_factor,
				data->max_iter, data->starty, data->skipy);
	return 0;
}

long JuliaWindow::thread_func_mandel(void *arg)
{
	thread_data *data = (thread_data *)arg;
	recalc_mandel(data->bits, data->xmod, data->size_x, data->size_y,
				data->xs, data->ys, data->x_factor, data->y_factor,
				data->max_iter, data->starty, data->skipy);
	return 0;
}


/*
 *  Bitmap view constructor
 */

BitmapView::BitmapView(BRect frame, BBitmap *bitmap)
	: BView(frame, "", B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS), the_bitmap(bitmap) {}


/*
 *  Blit the bitmap
 */

void BitmapView::Draw(BRect update)
{
	DrawBitmapAsync(the_bitmap, update, update);
}


/*
 *  View was resized, inform window
 */

void BitmapView::FrameResized(float width, float height)
{
	BMessage *msg = new BMessage(MSG_RESIZE);
	msg->AddLong("x", width);
	msg->AddLong("y", height);
	Window()->PostMessage(msg);
}


/*
 *  Mouse button pressed, send coordinates to window
 */

void BitmapView::MouseDown(BPoint point)
{
	ulong buttons = Window()->CurrentMessage()->FindLong("buttons");
	if (buttons & B_PRIMARY_MOUSE_BUTTON) {
		BMessage *msg = new BMessage(MSG_MOUSE_DOWN);
		msg->AddLong("x", point.x);
		msg->AddLong("y", point.y);
		Window()->PostMessage(msg);
	} else if (buttons & B_SECONDARY_MOUSE_BUTTON) {
		BMessage *msg = new BMessage(MSG_MOUSE_DOWN_2);
		msg->AddLong("x", point.x);
		msg->AddLong("y", point.y);
		Window()->PostMessage(msg);
	} else if (buttons & B_TERTIARY_MOUSE_BUTTON) {
		BMessage *msg = new BMessage(MSG_MOUSE_DOWN_3);
		msg->AddLong("x", point.x);
		msg->AddLong("y", point.y);
		Window()->PostMessage(msg);
	}
}


/*
 *  Mouse moved, send coordinates to application
 */

void BitmapView::MouseMoved(BPoint where, ulong code, BMessage *a_message)
{
	// Button down?
	ulong buttons;
	GetMouse(&where, &buttons, TRUE);
	where.ConstrainTo(Bounds());
	if ((buttons & B_PRIMARY_MOUSE_BUTTON) && Window()->IsActive()) {

		// Send coordinates
		BMessage *msg = new BMessage(MSG_MOUSE_MOVED);
		msg->AddLong("x", where.x);
		msg->AddLong("y", where.y);
		Window()->PostMessage(msg);
	}
}
